#!/usr/bin/python
"""Analyzes the package importer log files collecting various historical data.

This may need to be tweaked as the logger usage evolves.
"""

import datetime
import errno
import operator
import os
import re
import sys
import time

sys.path.insert(0, os.path.dirname(__file__))
import icommon


class LoggerLineMatcher(object):

    actor = None

    def __init__(self, actor, ident, regexp, args=None):
        """Logger line scanner helper.

        :param actor: Uniquely identify an actor in the log file.

        :param ident: Uniquely identify a logger line type for agiven ``actor``.

        :param regexp: Regular expression the message should match.

        :param args: A list a indices in the groups() matched, each
            corresponding group will be added in the order given in the
            details().
        """
        self.actor = actor
        self.ident = ident
        self.regexp = re.compile(regexp)
        self.args = args

    def matches(self, msg):
        self.match = self.regexp.match(msg)
        return self.match is not None


class ActorParser(object):
    """A parser for an actor defined by the lines it emits in a log file."""

    def __init__(self, ident):
        self.ident = ident
        self.estimate = False
        self.start = None # datetime.datetime object
        self.end = None # datetime.datetime object
        self.duration = None #datetime.timedelta object
        self.exception = None
        self.tb = None

    def see(self, log, state, stamp, args):
        """See a new state for the actor.

        :param log: The Log object for getting more lines if needed.

        :param state: The state seen.

        :param stamp: The datetime.datetime associated object.

        :param args: A state specific tuple.
        """
        raise NotImplementedError(self.see)

    def starts(self, stamp):
        if self.start is None:
            self.start = stamp

    def ends(self, stamp):
        if self.end is None:
            self.end = stamp

    def dead(self):
        return self.end is not None

    def round_time(self, t):
        return t - datetime.timedelta(microseconds=t.microsecond)

    def round_delta(self, d):
        return d - datetime.timedelta(microseconds=d.microseconds)

    def report_key(self):
        return (self.end, self.start)

    def report_times(self, out, epoch=None, apocalypse=None):
        if self.start is None:
            self.estimate = True
            self.start = epoch
        out.write('%s%s' % ([' ', '<'][self.estimate],
                                  self.round_time(self.start)))
        if self.end is None:
            self.estimate = True
            self.end = apocalypse
        duration = self.end - self.start
        if duration.seconds >= 10:
            # Lost in the noise
            duration = self.round_delta(duration)
        self.duration = duration
        out.write(' %s%s' % ([' ', '>'][self.estimate], duration))
        out.write(' %s\n' % (self.ident,))
        if self.exception is not None:
            exc_text = self.exception
            max_len = 60
            if len(exc_text) > max_len:
                # Naaah, too verbose
                closing = '...'
                exc_text = exc_text[:(max_len-len(closing))] + closing
            out.write('  Exception: %s\n' % (exc_text,))

    def parse_traceback(self, log):
        """Slurp all lines until the next valid logger line.

        The caller knows there is a traceback there, there is no strict check.

        :param log: A Log object.
        """
        something = []
        while True:
            lineno, line = log.next_line()
            # Trying to verify that ``line`` is a proper logger line is far too
            # expensive, so we try a cheap substitute instead the initial date
            # will start with 20 for the current century.
            if line.startswith('20'):
                log.pushed_back_line = line
                break
            if not line in ('', '\n'):
                something.append(line)
        self.tb = ''.join(something)
        if something:
            self.exception = something[-1].rstrip()
        else:
            self.exception = 'Exception with an empty traceback ???'


class ControllerParser(ActorParser):

    def __init__(self):
        super(ControllerParser, self).__init__('controller')
        self.nb_packages = 0
        self.cumulated_duration = datetime.timedelta(0)
        self.longest = None
        self.longest_duration = datetime.timedelta(0)

    def see(self, log, state, stamp, args):
        if state == 'starting':
            self.starts(stamp)
        elif state in ('signaled', 'waiting', 'mentioning', 'locked'):
            pass
        else:
            assert state == 'finished'
            self.ends(stamp)

    def package_ends(self, package_parser):
        if package_parser.duration > self.longest_duration:
            self.longest = package_parser.ident
            self.longest_duration = package_parser.duration
        self.cumulated_duration += package_parser.duration
        self.nb_packages += 1

    def report_times(self, out, epoch=None, apocalypse=None):
        super(ControllerParser, self).report_times(out, epoch, apocalypse)
        if self.nb_packages == 0:
            self.nb_packages = 1 # avoid divide by zero
        out.write('%s\n' % ('=' * 60,))
        out.write('  imported %6d, average %10s, longest: %10s %s\n'
                  % (self.nb_packages, self.round_delta(self.cumulated_duration
                                                        / self.nb_packages),
                     self.round_delta(self.longest_duration), self.longest,
                     ))
        out.write('%s\n' % ('=' * 60,))


class DriverParser(ActorParser):

    def __init__(self):
        super(DriverParser, self).__init__('driver')
        self.concurrency = 0

    def see(self, log, state, stamp, args):
        if state == 'starting':
            self.starts(stamp)
        elif state == 'active':
            packages = []
            for p in args[0].split(','):
                p = p.strip()
                if p[:2] == "u'" and p[-1] == "'":
                    p = p[2:-1]
                packages.append(p)
            concurrency = len(packages)
            if concurrency != self.concurrency:
                self.concurrency = concurrency
                print '>>> %s - Concurrency is now: %d' % (stamp, concurrency,)
            for p in packages:
                if p not in log.package_parsers:
                    # For partial logs, we may miss the package start
                    pp = log.get_package_parser(p)
                    pp.starts(log.last_active)
                    pp.estimate = True

            # There are ordering issues for Package log messages which means no
            # package log message is unambiguously the last one for a given
            # package. We use 'active' as one way to clearly decide when a
            # package is done.
            log.running_packages(packages)
            log.last_active = stamp
        elif state == 'failed':
            self.ends(stamp)
            # This is followed by "something" containing a
            # traceback
            self.parse_traceback(log)
        elif state in ('asked_to_stop', 'crashed', 'killing'):
            pass
        else:
            assert state in ('stopping', 'finished')
            self.ends(stamp)


class PackageParser(ActorParser):

    def __init__(self, name):
        super(PackageParser, self).__init__(name)
        self.tb = None

    def see(self, log, state, stamp, args):
        if state in ('trying', 'starting'):
            self.starts(stamp)
        elif state in ('locked', 'finished'):
            self.ends(stamp)
        elif state == 'success':
            # There is something to grab in args there, distro, pocket, etc
            self.ends(stamp)
        else:
            assert state == 'failed'
            self.ends(stamp)
            # This is followed by "something" containing a
            # traceback
            self.parse_traceback(log)


class Log(object):

    def __init__(self, path):
        self.path = path
        self.stream = iter(open(self.path))
        self.out = sys.stdout
        self.lineno = 0
        self.pushed_back_line = None
        # The first and last stamp seen in the log file
        self.epoch = None
        self.apocalypse = None
        # The last time a 'threads for [...] still active' was seen
        self.last_active = None
        # A logger line is:
        #  <time stamp> - <component?> - level - msg
        self.logger_line_re = re.compile(
            '([0-9- :,]{23}) - ([^ -]+) - ([A-Z]+) - (.*)')

        def Controller(ident, regexp):
            return LoggerLineMatcher('controller', ident, regexp)

        def Driver(ident, regexp):
            return LoggerLineMatcher('driver', ident, regexp)

        def Package(ident, regexp):
            return LoggerLineMatcher('package', ident, regexp)


        # All the possible lines, most often found first (as a cheap shot
        # optimization).
        self.matchers = [
            Package('trying', 'Trying ([^ ]+)'),
            Package('starting', 'Starting thread for ([^ ]+)'),
            Package('locked', "Couldn't lock ([^ ]+), skipping"),
            Package('success', 'Success ([^ ]+): (.*)'),
            Package('finished', 'thread for ([^ ]+) finished'),
            Package('failed', 'Importing ([^ ]+) failed:'),
            # Driver logger lines
            Driver('active', 'threads for \[([^\]]*)\] still active'),
            Driver('killing', 'Importing ([^ ]+) killed, timeout .* exceeded'),
            Driver('asked_to_stop', 'Driver asked to stop'),
            Driver('failed', 'Driver failed with exception:'),
            Driver('crashed', 'Driver hit (.*)'),
            Driver('stopping', 'Driver stopping'),
            Driver('finished', 'Driver finished: stopping'),
            # Controller logger lines
            Controller('starting', 'Starting up'),
            Controller('mentioning', '(PATH|CWD): (.*)'),
            Controller('locked', 'Another main process is running'),
            Controller('signaled', 'Received signal'),
            Controller('waiting', 'Waiting for driver to finish'),
            Controller('finished', 'Finished'),
            ]
        # At any given point in time, there could be one controller, one driver
        # and several packages writing to the log file. There is no guarantee
        # that any of them will have corresponding lines for starting or
        # finishing (not to mention dieing).
        self.controller = None
        self.driver = None
        self.package_parsers = {}

    def scan_logger_line(self, stamp_str, component, level, msg):
        """Scan a line already known to come from the logger.

        :param stamp_str: The leading date stamp.

        :param component: Always __main__ so far.

        :param level: The logger level INFO, DEBUG, etc.

        :param msg: The actual message at the end of the logged line.

        :returns: A (stamp, actor, ident, args) tuple where:
            - 'stamp' is the datetime.datetime object for the line leading 
              date stamp.
            - 'actor' is one of (controller, driver, package)
            - ident is an actor-specific state/action/whatever
            - args is an ident-specific tuple
        """
        stamp = datetime.datetime.strptime(stamp_str[:-4], '%Y-%m-%d %H:%M:%S')
        # strptime can't parse the miliseconds, do it manually
        stamp = stamp.replace(microsecond=int(stamp_str[-3:])*1000)
        # Initialize the stamp limits for the whole log file
        if self.epoch is None:
            self.epoch = stamp
        self.apocalypse = stamp

        for m in self.matchers:
            if m.matches(msg):
                return (stamp, m.actor, m.ident, m.match.groups())

        return (stamp, 'unknown', 'garbage', (stamp_str, component, level, msg))

    def next_line(self):
        """Get a new line from the current file.
        
        A pushed back line is handled for edge cases.
        """
        # We don't need to handle more than one pushed back line so far, a
        # stack could easily be implemented if needed -- vila 20110224
        if self.pushed_back_line is not None:
            # Give back the pushed back line without changing the lineno
            line = self.pushed_back_line
            self.pushed_back_line = None
        else:
            line = self.stream.next()
            self.lineno +=1
        return self.lineno, line

    def scan(self):
        """Iterate over all known lines or garbage."""
        while True:
            lineno, line = self.next_line()
            res = self.logger_line_re.match(line)
            if res is not None:
                stamp, component, level, msg = res.groups()
                yield lineno, 'logger', self.scan_logger_line(
                    stamp, component, level, msg)
            else:
                yield lineno, 'garbage', line

    def get_controller(self):
        """Get the current controller object, creating it if needed."""
        # autovivification
        if self.controller is None:
            self.controller = ControllerParser()
        return self.controller

    def get_driver(self):
        """Get the current driver object, creating it if needed."""
        # autovivification
        if self.driver is None:
            self.driver = DriverParser()
        return self.driver

    def get_package_parser(self, name):
        """Get the current package parser object, creating it if needed."""
        try:
            p = self.package_parsers[name]
        except KeyError:
            # autovivification
            p = PackageParser(name)
            self.package_parsers[name] = p
        return p

    def running_packages(self, packages):
        for name, pp in [(k,v) for k,v in self.package_parsers.items()]:
            if pp.dead() or name not in packages:
                pp.report_times(self.out, self.epoch, self.apocalypse)
                self.get_controller().package_ends(pp)
                del self.package_parsers[name]

    def report_all(self):
        """Report all data already gathered.

        This includes the controller, driver and packages data.
        """
        c = self.get_controller()
        try:
            packages = sorted(self.package_parsers.values(),
                              key=operator.methodcaller('report_key'))
        except TypeError:
            import pdb; pdb.set_trace()
        for p in packages:
            # Sorted by p.start (including None)
            p.report_times(self.out, self.epoch, self.apocalypse)
            c.package_ends(p)
        self.get_driver().report_times(self.out, self.epoch, self.apocalypse)
        c.report_times(self.out, self.epoch, self.apocalypse)

    def parse(self):
        """Parse a log file, outputting statistics about the package importer.
        """
        # controller and driver always run together (there is a bug if the
        # controller is killed with a hammer, we don't care).
        self.get_controller()
        self.get_driver()
        try:
            for lineno, typ, args in self.scan():
                if typ == 'logger':
                    (stamp, actor, ident,
                     remaining) = args[0], args[1], args[2], args[3]
                    if actor == 'controller':
                        c = self.get_controller()
                        c.see(self, ident, stamp, remaining)
                        d = self.get_driver()
                        if c.dead() or d.dead():
                            self.running_packages([])
                            self.report_all()
                            self.package_parsers = {}
                            self.controller = self.driver = None
                    elif actor == 'driver':
                        d = self.get_driver()
                        d.see(self, ident, stamp, remaining)
                    elif actor == 'package':
                        name = remaining[0]
                        p = self.get_package_parser(name)
                        p.see(self, ident, stamp, remaining[1:])
                    else:
                        # Unkown actor, same as garbage as far as reporting is
                        # concerned.
                        self.out.write('%s:%s garbage: %r\n'
                                         % (self.path, lineno, args))
                else:
                    line = args
                    if line != '\n': # Don't bother reporting empty lines
                        # Unrecognized line type
                        self.out.write('%s:%s garbage: %s'
                                         % (self.path, lineno, line,))
        finally:
            self.stream.close()
        # Pretends no more packages are running at this point
        self.running_packages([])
        # Final reports can break with log files without a single logger line
        # (no time stamps seen, we don't care)
        self.report_all()


def main(args):
    if len(args) != 2:
        print 'Usage: %s <log file path>'
        exit(1)

    log_path = args[1]
    if not os.path.exists(log_path):
        print '%s does not exist' % (log_path,)
        exit(1)

    log = Log(log_path)
    log.parse()


if __name__ == '__main__':
    main(sys.argv)
