#!/usr/bin/python

import codecs
import datetime
import logging
import os
import Queue
import random
import signal
from stat import ST_DEV, ST_INO
import subprocess
import sys
import threading
import time

from launchpadlib.errors import HTTPError

import icommon
import iconfig

#import httplib2
#httplib2.debuglevel = 1


class WatchedFileHandler(logging.FileHandler):
    """
    A handler for logging to a file, which watches the file
    to see if it has changed while in use. This can happen because of
    usage of programs such as newsyslog and logrotate which perform
    log file rotation. This handler, intended for use under Unix,
    watches the file to see if it has changed since the last emit.
    (A file has changed if its device or inode have changed.)
    If it has changed, the old file stream is closed, and the file
    opened to get a new stream.

    This handler is not appropriate for use under Windows, because
    under Windows open files cannot be moved or renamed - logging
    opens the files with exclusive locks - and so there is no need
    for such a handler. Furthermore, ST_INO is not supported under
    Windows; stat always returns zero for this value.

    This handler is based on a suggestion and patch by Chad J.
    Schroeder.
    """
    def __init__(self, filename, mode='a', encoding=None):
        logging.FileHandler.__init__(self, filename, mode, encoding)
        if not os.path.exists(self.baseFilename):
            self.dev, self.ino = -1, -1
        else:
            stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
        self.encoding = encoding

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        if self.encoding is None:
            stream = open(self.baseFilename, self.mode)
        else:
            stream = codecs.open(self.baseFilename, self.mode, self.encoding)
        return stream


    def emit(self, record):
        """
        Emit a record.

        First check if the underlying file has changed, and if it
        has, close the old stream and reopen the file to get the
        current stream.
        """
        if not os.path.exists(self.baseFilename):
            stat = None
            changed = 1
        else:
            stat = os.stat(self.baseFilename)
            changed = (stat[ST_DEV] != self.dev) or (stat[ST_INO] != self.ino)
        if changed and self.stream is not None:
            self.stream.close()
            self.stream = self._open()
            if stat is None:
                stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
        logging.FileHandler.emit(self, record)


def create_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    conf = iconfig.Iconfig()
    log_debug_path = conf.get('pkgimport.driver.log.debug')
    log_progress_path = conf.get('pkgimport.driver.log.progress')
    for p in (log_debug_path, log_progress_path):
        dir_path = os.path.dirname(p)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

    debug_handler = WatchedFileHandler(log_debug_path)
    progress_handler = WatchedFileHandler(log_progress_path)

    debug_handler.setLevel(logging.DEBUG)
    progress_handler.setLevel(logging.INFO)

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    debug_handler.setFormatter(formatter)
    progress_handler.setFormatter(formatter)

    logger.addHandler(debug_handler)
    logger.addHandler(progress_handler)
    return logger

logger = create_logger()
logger.info("Starting up")
# Some context to ease debug
logger.info("PATH: %s" % os.environ.get('PATH', None))
logger.info("CWD: %s" % os.getcwd())


def report_exception(component, exc):
    import traceback
    exc_class, exc_value, exc_tb = exc
    logger.info('%s failed with exception:\n%s'
                % (component,
                   ''.join(traceback.format_exception(exc_class, exc_value,
                                                      exc_tb))))

class Importer(icommon.SubprocessMonitor):

    def __init__(self, args, package_name, job_id):
        # FIXME: We should put the max_duration in a config file when it
        # becomes available. Until then, during the wheezy import the longest
        # import took more than 2 hours, so 24 is generous enough. On the other
        # hand, we had evidence of imports hanging for weeks (at least a month
        # even in one case, so 24 hours will get rid of them)
        super(Importer, self).__init__(args,
                                       # No more than a day
                                       max_duration=24*60*60)
        self.package_name = package_name
        self.job_id = job_id
        self.status_db = icommon.StatusDatabase(icommon.sqlite_file)

    def spawn(self):
        logger.info("Trying %s" % self.package_name)
        super(Importer, self).spawn()

    def collect(self, timeout=None):
        super(Importer, self).collect(timeout)
        if self.retcode == icommon.no_lock_returncode:
            logger.info("Couldn't lock %s, skipping" % self.package_name)
        else:
            # We mostly care about self.err which contains the traceback here
            output = self.out + self.err
            unicode_output = output.decode("utf-8", "replace")
            ascii_output = unicode_output.encode("ascii", "replace")
            success = self.retcode == 0
            if success:
                logger.info("Success %s: %s"
                            % (self.package_name,
                               ascii_output.replace("\n", " ")))
            else:
                logger.warning("Importing %s failed:\n%s" % (self.package_name,
                                                             ascii_output))
            self.status_db.finish_job(
                self.package_name, self.job_id, success,
                unicode_output.encode("utf-8", "replace"))
        logger.info("thread for %s finished" % self.package_name)


class AllQueue(Queue.Queue):
    """A Queue that always returns a package.

    The jobs in the status db are tried first. 

    When no more jobs are pending, the package db is used, even if we are not
    sure that it is needed.
    """

    def __init__(self):
        Queue.Queue.__init__(self)
        self.tried = set()
        self.packages_db = icommon.PackageDatabase(icommon.sqlite_package_file)
        self.status_db = icommon.StatusDatabase(icommon.sqlite_file)

    def get_nowait(self):
        # FIXME: It's bogus to implement get_nowait() this way (without
        # locking) but we don't really care since we have a single consumer
        # and no producer. This should be changed to a different model anyway
        # so different kinds of jobs could be queued -- vila 20110221
        try:
            job_id, package_name = Queue.Queue.get_nowait(self)
        except Queue.Empty:
            self.unfinished_tasks += 1
            job_id, package_name = self.next_job()
        # FIXME: Gimme a config option for that, kthksbye -- vila 20110224
        script_path = '%sscripts/import_package.py' % (os.environ['BASEDIR'],)
        return Importer([script_path, package_name],
                        package_name, job_id)

    def next_job(self):
        while True:
            # jobs first
            job_id, package = self.status_db.next_job()
            if package is not None:
                return job_id, package
            else:
                # Now the packages
                try:
                    to_try = self.packages_db.get_one(self.tried)
                except StopIteration:
                    raise Queue.Empty
                if to_try is not None:
                    self.tried.add(to_try)
                    job_id = self.status_db.start_package(to_try)
                    if job_id is not None:
                        return (job_id, to_try)
                else:
                    logger.info("All packages requeued, start again")
                    self.tried = set()


class ImportDriver(icommon.ThreadDriver):
    """Monitor the ThreadedImporter.

    This includes planning and spawning imports, tracking their success or
    failures and shutting them down.
    """

    def __init__(self, max_threads):
        super(ImportDriver, self).__init__(None, max_threads)

    def before_start(self):
        super(ImportDriver, self).before_start()
        self.queue = AllQueue()

    def do_one_step(self):
        try:
            super(ImportDriver, self).do_one_step()
        except Exception, e:
            # The Driver must go on !
            report_exception('Driver', sys.exc_info())

    def before_stop(self):
        try:
            super(ImportDriver, self).before_stop()
        except Exception, e:
            # The Driver must go on !
            report_exception('Driver', sys.exc_info())

    def start_new_thread(self):
        t = super(ImportDriver, self).start_new_thread()
        if t is not None:
            logger.info("Starting thread for %s" % (t.package_name,))

    def collect_terminated_threads(self):
        before = len(self.threads)
        super(ImportDriver, self).collect_terminated_threads()
        after = len(self.threads)
        if before != after:
            # Only mention the running threads when we add or remove some
            logger.info("threads for %r still active"
                        % [t.package_name for t in self.threads])

    def check_time_quota(self):
        killed = super(ImportDriver, self).check_time_quota()
        for t in killed:
            logger.warning("Importing %s killed, timeout %s exceeded\n"
                           % (t.package_name,
                              datetime.timedelta(seconds=t.max_duration)))

class Stop(Exception):
    pass


class ImportController(object):
    """Monitor the ImportDriver and interpret signals."""

    def __init__(self):
        self.status_db = icommon.StatusDatabase(icommon.sqlite_file)
        # Start with an 8-way Driver
        self.driver = ImportDriver(8)

    def run(self):
        lock = icommon.lock_main()
        if lock is None:
            logger.info("Another main process is running")
            raise Stop
        try:
            # First, re-queue previously interrupted jobs.
            self.status_db.add_jobs_for_interrupted()
            self.driver.start()
            self.driver.started.wait()
            while True:
                # First, check the driver. We don't check more often than the
                # timeout given below. This timeout should be a balance between
                # user expectations and resources usage, there is no point
                # doing all the filesystem calls every milisecond but the user
                # shouldn't have to wait too long to see his modifications
                # taken into account either.
                self.driver.stopped.wait(0.2)
                if self.driver.stopped.isSet():
                    break
                # Then check the graceful stop file
                if (os.path.exists(icommon.stop_file)
                    and not self.driver.queue_closed.isSet()):
                    self.driver.queue_closed.set()
                    logger.info('Driver would not process new requests')
                    continue
                # And the number of threads
                nb_threads = self.get_max_threads()
                if self.driver.max_threads != nb_threads:
                    logger.info('Read %d in %s'
                                % (nb_threads, icommon.max_threads_file))
                    self.driver.max_threads = nb_threads
            # Catch driver exception if any
            self.report_driver_exception()
            self.driver.join()
            logger.info("Finished")
        finally:
            lock.close()

    def get_max_threads(self):
        # By default, keep the existing one
        nb_threads = self.driver.max_threads
        if os.path.exists(icommon.max_threads_file):
            f = open(icommon.max_threads_file)
            try:
                line = f.readline().strip()
                if line:
                   nb_threads = int(line)
            except Exception, e:
                logger.warning('Error reading max threads file %s: %s'
                               % (icommon.max_threads_file, str(e)))
            finally:
                f.close()
        return nb_threads

    def report_driver_exception(self):
        if self.driver.exception is not None:
            report_exception('Driver', self.driver.exception)
            # No need to ever re-raise this exception, we just reported it
            self.driver.exception = None


controller = ImportController()


def die(signum, frame):
    logger.info("Received signal")
    signal.signal(signum, signal.SIG_DFL)

    controller.driver.stop.set()
    logger.info("Waiting for driver to finish")
    controller.driver.stopped.wait()
    controller.report_driver_exception()
    controller.driver.join()
    logger.info("Driver finished: stopping")
    sys.exit(1)

# Install signal handlers
signal.signal(signal.SIGINT, die)
signal.signal(signal.SIGTERM, die)


# Run the controller
try:
    controller.run()
except HTTPError, e:
    logger.critical(e.content)
    raise
except Exception, e:
    logger.critical(str(e))
    raise
