import os

from bzrlib import (
    tests,
    transform,
    )


import import_package


class TestCheckSame(tests.TestCaseWithTransport):

    def assertTreesEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertTrue(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are not actually equivalent')

    def assertTreesNotEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertFalse(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are accidentally equivalent')

    def test_identical_different_revs_same_fids(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/file', 'content\n')])
        t1.add(['file'], ['file-id'])
        revid = t1.commit('first')
        t2 = t1.bzrdir.sprout('tree2').open_workingtree()
        # Both trees independently get identical content
        self.build_tree_contents([('tree1/file', 'new content\n'),
                                  ('tree2/file', 'new content\n')])
        t1_rev = t1.commit('new content')
        t2_rev = t2.commit('new content')
        t1.branch.repository.fetch(t2.branch.repository, t2_rev)

        self.assertTreesEquivalent(repo, t1_rev, t2_rev)

    def test_exec_patch_change_is_ignored(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/patches/',),
                                  ('tree1/debian/patches/my.patch', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/patches', 'debian/patches/my.patch'])
        rev1 = t1.commit('create the new tree')
        tt = transform.TreeTransform(t1)
        tt.set_executability(True,
            tt.trans_id_file_id(t1.path2id('debian/patches/my.patch')))
        tt.apply()
        rev2 = t1.commit('set executable bit')
        self.assertTreesEquivalent(repo, rev1, rev2)
        self.build_tree_contents([
            ('tree1/debian/patches/my.patch', 'new content\n'),])
        rev3 = t1.commit('change content')
        self.assertTreesNotEquivalent(repo, rev1, rev3)
        self.assertTreesNotEquivalent(repo, rev2, rev3)

    def test_deleting_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ('tree1/debian/dir/file2', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file', 'debian/dir/file2'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file2')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_last_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_cascaded_dirs_is_not_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/subdir1/',),
                                  ('tree1/debian/dir/subdir2/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/subdir1',
                'debian/dir/subdir2'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/debian/dir/subdir1')
        os.rmdir('tree1/debian/dir/subdir2')
        os.rmdir('tree1/debian/dir')
        rev2 = t1.commit('del subtree')
        self.assertTreesEquivalent(repo, rev1, rev2)
        # The reverse change is *not* equivalent
        self.assertTreesNotEquivalent(repo, rev2, rev1)

    def test_deleting_non_debian_empty_dir_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/subdir1/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'subdir1'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/subdir1')
        rev2 = t1.commit('del subbdir')
        self.assertTreesNotEquivalent(repo, rev1, rev2)


