import os
import Queue
import signal
import threading
import time


from bzrlib import tests


import icommon


base_dir = os.path.abspath(os.path.dirname(icommon.__file__))
def get_test_script_path(name):
    return os.path.join(base_dir, 'tests', name)


class ExceptionForTests(Exception):
    """This is intended to be used by tests to identify a specific exception.

    This allows test to catch this exception instead of a possibly ambiguous
    one.
    """
    pass


class TestSubprocessMonitor(tests.TestCase):

    def run_in_subprocess(self, args, klass=None):
        if klass is None:
            klass = icommon.SubprocessMonitor
        sub = klass(args)
        self.addCleanup(sub.join, 0)
        sub.start()
        return sub

    def run_sleeper_in_subprocess(self, klass=None):
        """Run a sleeper in a subprocess.

        This is used to simulate a never-ending process without consuming too
        much CPU cycles.
        """
        cmd = 'import time; while True: time.sleep(0.001)'
        return self.run_in_subprocess(['-c', cmd], klass=klass)

    def test_broken_command(self):
        sub = self.run_in_subprocess(['this-is-not-a-python-script'])
        sub.stopped.wait()
        self.assertEquals(True, sub.started.isSet())
        self.assertEquals(
            "python: can't open file 'this-is-not-a-python-script':"
            " [Errno 2] No such file or directory\n",
            sub.err)

    def test_success(self):
        sub = self.run_in_subprocess(['-c', 'print "toto"'])
        sub.stopped.wait()
        self.assertEquals('toto\n', sub.out)
        self.assertEquals('', sub.err)
        self.assertEquals(0, sub.retcode)

    def test_failure(self):
        sub = self.run_in_subprocess(['-c', 'import sys; sys.exit(1)'])
        sub.stopped.wait()
        self.assertEquals('', sub.out)
        self.assertEquals('', sub.err)
        self.assertEquals(1, sub.retcode)

    def test_kill(self):
        sub = self.run_sleeper_in_subprocess()
        sub.started.wait()
        self.assertEquals(True, sub.kill_with_escalation())
        sub.stopped.wait()
        self.assertEquals(-signal.SIGTERM, sub.retcode)

    def test_kill_SIGKILL(self):
        sub = self.run_sleeper_in_subprocess()
        sub.started.wait()
        self.assertEquals(True, sub._kill(signal.SIGKILL))
        sub.stopped.wait()
        self.assertEquals(-signal.SIGKILL, sub.retcode)

    def test_kill_catch_zombie(self):
        kill_me = threading.Event()
        resume = threading.Event()
        class TestMonitor(icommon.SubprocessMonitor):

            def switch_and_set(self, new):
                super(TestMonitor, self).switch_and_set(new)
                if new is self.stopped:
                    # The process is running but we haven't called
                    # proc.communicate yet.
                    kill_me.set()
                    resume.wait()

        sub = self.run_sleeper_in_subprocess(klass=TestMonitor)
        self.addCleanup(sub.join, 0)
        sub.started.wait()
        kill_me.wait()
        # Kill the subprocess ourselves
        os.kill(sub.proc_pid, signal.SIGTERM)
        pid, st = os.waitpid(sub.proc_pid, 0)
        self.assertEquals(sub.proc_pid, pid)
        # Only the low byte contains the signal and the high bit if for core
        # dumps.
        self.assertEquals(signal.SIGTERM, st & 0x7F)
        # Release the thread so it can call proc.communicate
        resume.set()
        # Now pretend we don't know the subprocess is dead
        self.assertEquals(False, sub.kill_with_escalation())
        sub.join()
        # No useful retcode available
        self.assertEquals(None, sub.retcode)
        # Nothing either in stdout/stderr. This means that subprocesses should
        # handled SIGTERM to allow us to collect their data.
        self.assertIs(None, sub.out)
        self.assertIs(None, sub.err)


class TestSubprocessMonitorWithScript(tests.TestCaseInTempDir):

    def run_in_subprocess(self, args, klass=None):
        if klass is None:
            klass = icommon.SubprocessMonitor
        sub = klass(args)
        self.addCleanup(sub.join, 0)
        sub.start()
        return sub

    def wait_for_file(self, path, timeout=2.0, a_slice=0.001):
        """Wait for the given file to appear on the file system.

        This is used as a simple inter-process synchronization mechanism.
        """
        waited_time = 0.0
        while waited_time < timeout:
            try:
                f = open(path)
                f.close()
                break
            except:
                time.sleep(a_slice)
                waited_time += a_slice
        else:
            self.fail('%s never appeared' % (path,))

    def test_kill_caught(self):
        sub = self.run_in_subprocess(
            [get_test_script_path('die_on_sigterm.py')])
        sub.started.wait()
        self.wait_for_file('ready_to_die')
        self.assertEquals(True, sub.kill_with_escalation())
        sub.stopped.wait()
        # The signal has been caught, we just get a failure return code, no
        # funky stuff with the signal in the low byte and so on
        self.assertEquals(1, sub.retcode)
        self.assertEndsWith(sub.err.splitlines()[-1], "I'm so dead")

    def test_kill_harder_directly(self):
        sub = self.run_in_subprocess(
            [get_test_script_path('survive_sigterm.py')])
        sub.started.wait()
        self.wait_for_file('ready_to_die')
        self.assertEquals(True, sub.kill_with_escalation())
        self.wait_for_file('surviving')
        # The subprocess is still alive !
        self.assertEquals(True, sub._kill(signal.SIGKILL))
        sub.stopped.wait()
        # The SIGKILL was the succeeding one
        self.assertEquals(-9, sub.retcode)
        self.assertEquals(sub.out, 'Try harder\n')

    def test_kill_survivor(self):
        sub = self.run_in_subprocess(
            [get_test_script_path('survive_sigterm.py')])
        sub.started.wait()
        self.wait_for_file('ready_to_die')
        self.assertEquals(True, sub.kill_with_escalation())
        self.wait_for_file('surviving')
        # The subprocess is still alive !
        self.assertEquals(False, sub.kill_with_escalation(grace_period=1.0))
        # Ok, we tried to soon, we'll be less patient now
        self.assertEquals(True, sub.kill_with_escalation(grace_period=0.0))
        sub.stopped.wait()
        # The SIGKILL was the succeeding one
        self.assertEquals(-9, sub.retcode)
        self.assertEquals(sub.out, 'Try harder\n')


class Sleeper(icommon.SubprocessMonitor):

    def __init__(self, sleep_time=0.001, max_duration=None):
        # sleep can't be killed (or interrupted) so we don't sleep for long but
        # keep sleeping until killed
        super(Sleeper, self).__init__(
            # Make sure we die long after the test finish (to avoid leaks) but
            # still use a value that allow *some* debug (you want more for
            # *heavy* debug :)
            ['-c', 'import time; while xrange(120.0/%s): time.sleep(%s)'
             % (sleep_time, sleep_time)],
            max_duration=max_duration)


class TestThreadDriver(tests.TestCase):

    def get_driver(self, *args, **kwargs):
        klass = kwargs.pop('klass', None)
        if klass is None:
            klass = icommon.ThreadDriver
        return klass(*args, **kwargs)

    def get_started_driver(self, target=None, override=None, queue=None,
                           max_threads=None, klass=None):
        if queue is None:
            # Creating a queue here roughly means nobody will ever touch it, so
            # it will stay empty. This allows tests about the behaviour of the
            # driver when no sub-threads are involved.
            queue = Queue.Queue(1)
        if max_threads is None:
            # Tests that don't specify max_threads don't expect threads at all
            max_threads = 0
        driver = self.get_driver(queue, max_threads, target=target, klass=klass)
        if override is not None:
            for meth_name, callback in override:
                setattr(driver, meth_name, callback)
        self.addCleanup(driver.join, 0)
        driver.start()
        return driver

    def test_start_stop(self):
        """A do-nothing driver can be started and stopped."""
        driver = self.get_started_driver()
        driver.started.wait()
        driver.stop.set()

    def test_broken_target(self):
        """Exception while starting is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(target=raise_my_exception)
        driver.started.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_to_start(self):
        """Exception in the starting hook is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('before_start', raise_my_exception)])
        driver.started.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_during_first_step(self):
        """Exception during the first single step is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('do_one_step', raise_my_exception)])
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_during_second_step(self):
        """Exception during the next steps is caught."""
        self.steps = 0
        def raise_my_exception():
            if self.steps > 1:
                raise ExceptionForTests()
            else:
                self.steps += 1
        driver = self.get_started_driver(
            override=[('do_one_step', raise_my_exception)])
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_while_stopping(self):
        """Exception during shutdown is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('before_stop', raise_my_exception)])
        driver.stop.set()
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_one_thread_finish(self):
        self.started = threading.Event()
        self.resumed = threading.Event()
        self.terminated = threading.Event()
        class TestThread(icommon.SubprocessMonitor):

            def collect(thread):
                self.assertIs(self.thread, thread)
                self.assertEquals(1, thread.retcode)
                self.terminated.set()
        self.thread = TestThread(['-c', 'import sys; sys.exit(1)'])
        self.addCleanup(self.thread.join, 0)
        q = Queue.Queue(1)
        q.put(self.thread)

        class TestDriver(icommon.ThreadDriver):

            def start_new_thread(driver):
                t = super(TestDriver, driver).start_new_thread()
                # all calls to start_new_thread except one will hit an empty
                # queue
                if t is not None:
                    self.assertIs(self.thread, t)
                    self.started.set()
                    self.resumed.wait()

        driver = TestDriver(queue=q, max_threads=1)
        self.addCleanup(driver.join, 0)
        driver.start()
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        # Wait for the subprocess thread to start
        self.started.wait()
        # the thread is started and at most wait for collection on
        # self.terminated so it's still there
        self.assertLength(1, driver.threads)
        self.assertEquals(self.thread, driver.threads[0])
        self.resumed.set()
        self.terminated.wait()
        driver.stop.set()
        driver.stopped.wait()
        self.assertLength(0, driver.threads)

    def test_thread_exception_raised_in_driver(self):
        class TestThread(icommon.SubprocessMonitor):

            def collect(thread):
                self.assertEquals(1, thread.retcode)
                raise ExceptionForTests()
        q = Queue.Queue(1)
        t = TestThread(['-c', 'import sys; sys.exit(1)'])
        self.addCleanup(t.join, 0)
        q.put(t)
        driver = self.get_started_driver(queue=q, max_threads=1)
        t.stopped.wait()
        driver.stop.set()
        self.assertRaises(ExceptionForTests, driver.join)

    def test_kill_pending_threads(self):
        q = Queue.Queue(3)
        for i in range(3):
            s = Sleeper()
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill_with_escalation)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=3)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        # Wait for the driver to empty the queue
        q.join()
        self.assertLength(3, driver.threads)
        # Tell the driver to stop *now*: kill the threads
        driver.stop.set()
        driver.stopped.wait()
        self.assertLength(0, driver.threads)

    def test_respect_max_threads(self):
        q = Queue.Queue(3)
        driver = self.get_started_driver(queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        self.assertLength(0, driver.threads)
        sleepers = [Sleeper() for _ in range(3)]
        for s in sleepers:
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill_with_escalation)
            q.put(s)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        self.assertLength(2, driver.threads)
        self.assertEquals([sleepers[0], sleepers[1]], driver.threads)
        # sleepers[2] is still... awake
        self.assertFalse(sleepers[2].started.isSet())
        sleepers[0].kill_with_escalation()
        sleepers[0].stopped.wait()
        # Now the last sleeper will be started
        sleepers[2].started.wait()
        self.assertLength(2, driver.threads)
        self.assertEquals([sleepers[1], sleepers[2]], driver.threads)

    def test_respect_lowering_max_threads(self):
        q = Queue.Queue(4)
        sleepers = [Sleeper() for _ in range(4)]
        for s in sleepers:
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill_with_escalation)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=3)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        sleepers[2].started.wait()
        self.assertLength(3, driver.threads)
        self.assertEquals([sleepers[0], sleepers[1], sleepers[2]],
                          driver.threads)
        driver.max_threads = 2
        # Kill 2 sleepers and wait
        sleepers[0].kill_with_escalation()
        sleepers[0].stopped.wait()
        sleepers[1].kill_with_escalation()
        sleepers[1].stopped.wait()
        # Wait for another sleeper to start (because it's easier to sync on a
        # sleeper start than trying to catch when a sleeper goes out of the
        # driver's scope)
        sleepers[3].started.wait()
        self.assertEquals([sleepers[2], sleepers[3]], driver.threads)

    def test_closing_queue_stop(self):
        q = Queue.Queue(3)
        sleepers = [Sleeper() for _ in range(3)]
        for s in sleepers[:-1]: # the third sleeper will never be run
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill_with_escalation)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        # close the queue
        driver.queue_closed.set()
        sleepers[0].kill_with_escalation()
        sleepers[1].kill_with_escalation()
        driver.stopped.wait()

    def test_check_quotas(self):
        # A sleeper that will be killed
        sleeper = Sleeper(sleep_time=1, max_duration=0.0)
        q = Queue.Queue(1)
        self.addCleanup(sleeper.join, 0)
        self.addCleanup(sleeper.kill_with_escalation)
        q.put(sleeper)
        self.killed = None
        class TestDriver(icommon.ThreadDriver):

            def check_time_quota(driver):
                killed = super(TestDriver, driver).check_time_quota()
                if killed:
                    self.killed = killed
        driver = self.get_started_driver(klass=TestDriver,
                                         queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        self.assertLength(0, driver.threads)
        sleeper.started.wait()
        # wait longer than the allowed duration
        time.sleep(0.01)
        self.assertEquals([sleeper], self.killed)
        driver.stop.set()
        driver.stopped.wait()


# FIXME: The ImportController is a bit harder to test as it handles signals and
# as such requires being run in the main thread.
class TestController(tests.TestCaseInTempDir):

    pass

